@@ -1,6 +1,6 @@
 
 /* Return the initial module search path. */
-/* Used by DOS, OS/2, Windows 3.1, Windows 95/98, Windows NT. */
+/* Used by DOS, Windows 3.1, Windows 95/98, Windows NT. */
 
 /* ----------------------------------------------------------------
    PATH RULES FOR WINDOWS:
@@ -25,7 +25,7 @@
      is set, we believe it.  Otherwise, we use the path of our host .EXE's
      to try and locate our "landmark" (lib\\os.py) and deduce our home.
      - If we DO have a Python Home: The relevant sub-directories (Lib,
-       plat-win, lib-tk, etc) are based on the Python Home
+       plat-win, etc) are based on the Python Home
      - If we DO NOT have a Python Home, the core Python Path is
        loaded from the registry.  This is the main PythonPath key,
        and both HKLM and HKCU are combined to form the path)
@@ -51,15 +51,18 @@
     exe, some very strange installation setup) you get a path with
     some default, but relative, paths.
 
+  * An embedding application can use Py_SetPath() to override all of
+    these authomatic path computations.
+
    ---------------------------------------------------------------- */
 
 
 #include "Python.h"
 #include "osdefs.h"
+#include <wchar.h>
 
 #ifdef MS_WINDOWS
 #include <windows.h>
-#include <tchar.h>
 #endif
 
 #ifdef HAVE_SYS_TYPES_H
@@ -79,20 +82,22 @@
  * The approach is an adaptation for Windows of the strategy used in
  * ../Modules/getpath.c; it uses the Windows Registry as one of its
  * information sources.
+ *
+ * Py_SetPath() can be used to override this mechanism.  Call Py_SetPath
+ * with a semicolon separated path prior to calling Py_Initialize.
  */
 
 #ifndef LANDMARK
-#define LANDMARK "lib\\os.py"
+#define LANDMARK L"lib\\os.py"
 #endif
 
-static char prefix[MAXPATHLEN+1];
-static char progpath[MAXPATHLEN+1];
-static char dllpath[MAXPATHLEN+1];
-static char *module_search_path = NULL;
+static wchar_t prefix[MAXPATHLEN+1];
+static wchar_t progpath[MAXPATHLEN+1];
+static wchar_t *module_search_path = NULL;
 
 
 static int
-is_sep(char ch) /* determine if "ch" is a separator character */
+is_sep(wchar_t ch)      /* determine if "ch" is a separator character */
 {
 #ifdef ALTSEP
     return ch == SEP || ch == ALTSEP;
@@ -105,36 +110,85 @@ is_sep(char ch) /* determine if "ch" is a separator character */
    beyond existing terminator.
 */
 static void
-reduce(char *dir)
+reduce(wchar_t *dir)
 {
-    size_t i = strlen(dir);
+    size_t i = wcsnlen_s(dir, MAXPATHLEN+1);
+    if (i >= MAXPATHLEN+1)
+        Py_FatalError("buffer overflow in getpathp.c's reduce()");
+
     while (i > 0 && !is_sep(dir[i]))
         --i;
     dir[i] = '\0';
 }
 
+static int
+change_ext(wchar_t *dest, const wchar_t *src, const wchar_t *ext)
+{
+    if (src && src != dest) {
+        size_t src_len = wcsnlen_s(src, MAXPATHLEN+1);
+        size_t i = src_len;
+        if (i >= MAXPATHLEN+1) {
+            Py_FatalError("buffer overflow in getpathp.c's reduce()");
+        }
+
+        while (i > 0 && src[i] != '.' && !is_sep(src[i]))
+            --i;
+
+        if (i == 0) {
+            dest[0] = '\0';
+            return -1;
+        }
+
+        if (is_sep(src[i])) {
+            i = src_len;
+        }
+
+        if (wcsncpy_s(dest, MAXPATHLEN+1, src, i)) {
+            dest[0] = '\0';
+            return -1;
+        }
+    } else {
+        wchar_t *s = wcsrchr(dest, L'.');
+        if (s) {
+            s[0] = '\0';
+        }
+    }
+
+    if (wcscat_s(dest, MAXPATHLEN+1, ext)) {
+        dest[0] = '\0';
+        return -1;
+    }
+
+    return 0;
+}
 
 static int
-exists(char *filename)
+exists(wchar_t *filename)
 {
-    struct stat buf;
-    return stat(filename, &buf) == 0;
+    return GetFileAttributesW(filename) != 0xFFFFFFFF;
 }
 
 /* Assumes 'filename' MAXPATHLEN+1 bytes long -
    may extend 'filename' by one character.
 */
 static int
-ismodule(char *filename)        /* Is module -- check for .pyc/.pyo too */
+ismodule(wchar_t *filename, int update_filename) /* Is module -- check for .pyc too */
 {
+    int n;
+
     if (exists(filename))
         return 1;
 
     /* Check for the compiled version of prefix. */
-    if (strlen(filename) < MAXPATHLEN) {
-        strcat(filename, Py_OptimizeFlag ? "o" : "c");
-        if (exists(filename))
-            return 1;
+    n = wcsnlen_s(filename, MAXPATHLEN+1);
+    if (n < MAXPATHLEN) {
+        int exist = 0;
+        filename[n] = L'c';
+        filename[n + 1] = L'\0';
+        exist = exists(filename);
+        if (!update_filename)
+            filename[n] = L'\0';
+        return exist;
     }
     return 0;
 }
@@ -149,23 +203,23 @@ ismodule(char *filename)        /* Is module -- check for .pyc/.pyo too */
    stuff as fits will be appended.
 */
 static void
-join(char *buffer, char *stuff)
+join(wchar_t *buffer, const wchar_t *stuff)
 {
-    size_t n, k;
-    if (is_sep(stuff[0]))
-        n = 0;
-    else {
-        n = strlen(buffer);
-        if (n > 0 && !is_sep(buffer[n-1]) && n < MAXPATHLEN)
-            buffer[n++] = SEP;
+    size_t n;
+    if (is_sep(stuff[0]) ||
+        (wcsnlen_s(stuff, 4) >= 3 && stuff[1] == ':' && is_sep(stuff[2]))) {
+        if (wcscpy_s(buffer, MAXPATHLEN+1, stuff) != 0)
+            Py_FatalError("buffer overflow in getpathp.c's join()");
+        return;
     }
-    if (n > MAXPATHLEN)
-        Py_FatalError("buffer overflow in getpathp.c's joinpath()");
-    k = strlen(stuff);
-    if (n + k > MAXPATHLEN)
-        k = MAXPATHLEN - n;
-    strncpy(buffer+n, stuff, k);
-    buffer[n+k] = '\0';
+
+    n = wcsnlen_s(buffer, MAXPATHLEN+1);
+    if (n > 0 && !is_sep(buffer[n - 1]) && n < MAXPATHLEN) {
+        buffer[n] = SEP;
+        buffer[n + 1] = '\0';
+    }
+    if (wcscat_s(buffer, MAXPATHLEN+1, stuff) != 0)
+        Py_FatalError("buffer overflow in getpathp.c's join()");
 }
 
 /* gotlandmark only called by search_for_prefix, which ensures
@@ -173,14 +227,13 @@ join(char *buffer, char *stuff)
    'landmark' can not overflow prefix if too long.
 */
 static int
-gotlandmark(char *landmark)
+gotlandmark(wchar_t *landmark)
 {
     int ok;
-    Py_ssize_t n;
+    Py_ssize_t n = wcsnlen_s(prefix, MAXPATHLEN);
 
-    n = strlen(prefix);
     join(prefix, landmark);
-    ok = ismodule(prefix);
+    ok = ismodule(prefix, FALSE);
     prefix[n] = '\0';
     return ok;
 }
@@ -188,10 +241,10 @@ gotlandmark(char *landmark)
 /* assumes argv0_path is MAXPATHLEN+1 bytes long, already \0 term'd.
    assumption provided by only caller, calculate_path() */
 static int
-search_for_prefix(char *argv0_path, char *landmark)
+search_for_prefix(wchar_t *argv0_path, wchar_t *landmark)
 {
     /* Search from argv0_path, until landmark is found */
-    strcpy(prefix, argv0_path);
+    wcscpy_s(prefix, MAXPATHLEN + 1, argv0_path);
     do {
         if (gotlandmark(landmark))
             return 1;
@@ -200,6 +253,20 @@ search_for_prefix(char *argv0_path, char *landmark)
     return 0;
 }
 
+
+static int
+get_dllpath(wchar_t *dllpath)
+{
+#ifdef Py_ENABLE_SHARED
+    extern HANDLE PyWin_DLLhModule;
+    if (PyWin_DLLhModule && GetModuleFileNameW(PyWin_DLLhModule, dllpath, MAXPATHLEN)) {
+        return 0;
+    }
+#endif
+    return -1;
+}
+
+
 #ifdef MS_WINDOWS
 #ifdef Py_ENABLE_SHARED
 
@@ -220,39 +287,40 @@ extern const char *PyWin_DLLVersionString;
    in advance.  It could be simplied now Win16/Win32s is dead!
 */
 
-static char *
+static wchar_t *
 getpythonregpath(HKEY keyBase, int skipcore)
 {
     HKEY newKey = 0;
     DWORD dataSize = 0;
     DWORD numKeys = 0;
     LONG rc;
-    char *retval = NULL;
-    TCHAR *dataBuf = NULL;
-    static const TCHAR keyPrefix[] = _T("Software\\Python\\PythonCore\\");
-    static const TCHAR keySuffix[] = _T("\\PythonPath");
-    size_t versionLen;
+    wchar_t *retval = NULL;
+    WCHAR *dataBuf = NULL;
+    static const WCHAR keyPrefix[] = L"Software\\Python\\PythonCore\\";
+    static const WCHAR keySuffix[] = L"\\PythonPath";
+    size_t versionLen, keyBufLen;
     DWORD index;
-    TCHAR *keyBuf = NULL;
-    TCHAR *keyBufPtr;
-    TCHAR **ppPaths = NULL;
+    WCHAR *keyBuf = NULL;
+    WCHAR *keyBufPtr;
+    WCHAR **ppPaths = NULL;
 
     /* Tried to use sysget("winver") but here is too early :-( */
-    versionLen = _tcslen(PyWin_DLLVersionString);
+    versionLen = strlen(PyWin_DLLVersionString);
     /* Space for all the chars, plus one \0 */
-    keyBuf = keyBufPtr = malloc(sizeof(keyPrefix) +
-                                sizeof(TCHAR)*(versionLen-1) +
-                                sizeof(keySuffix));
+    keyBufLen = sizeof(keyPrefix) +
+                sizeof(WCHAR)*(versionLen-1) +
+                sizeof(keySuffix);
+    keyBuf = keyBufPtr = PyMem_RawMalloc(keyBufLen);
     if (keyBuf==NULL) goto done;
 
-    memcpy(keyBufPtr, keyPrefix, sizeof(keyPrefix)-sizeof(TCHAR));
-    keyBufPtr += sizeof(keyPrefix)/sizeof(TCHAR) - 1;
-    memcpy(keyBufPtr, PyWin_DLLVersionString, versionLen * sizeof(TCHAR));
+    memcpy_s(keyBufPtr, keyBufLen, keyPrefix, sizeof(keyPrefix)-sizeof(WCHAR));
+    keyBufPtr += Py_ARRAY_LENGTH(keyPrefix) - 1;
+    mbstowcs(keyBufPtr, PyWin_DLLVersionString, versionLen);
     keyBufPtr += versionLen;
     /* NULL comes with this one! */
     memcpy(keyBufPtr, keySuffix, sizeof(keySuffix));
     /* Open the root Python key */
-    rc=RegOpenKeyEx(keyBase,
+    rc=RegOpenKeyExW(keyBase,
                     keyBuf, /* subkey */
             0, /* reserved */
             KEY_READ,
@@ -266,31 +334,31 @@ getpythonregpath(HKEY keyBase, int skipcore)
     /* Allocate a temp array of char buffers, so we only need to loop
        reading the registry once
     */
-    ppPaths = malloc( sizeof(TCHAR *) * numKeys );
+    ppPaths = PyMem_RawMalloc( sizeof(WCHAR *) * numKeys );
     if (ppPaths==NULL) goto done;
-    memset(ppPaths, 0, sizeof(TCHAR *) * numKeys);
+    memset(ppPaths, 0, sizeof(WCHAR *) * numKeys);
     /* Loop over all subkeys, allocating a temp sub-buffer. */
     for(index=0;index<numKeys;index++) {
-        TCHAR keyBuf[MAX_PATH+1];
+        WCHAR keyBuf[MAX_PATH+1];
         HKEY subKey = 0;
         DWORD reqdSize = MAX_PATH+1;
         /* Get the sub-key name */
-        DWORD rc = RegEnumKeyEx(newKey, index, keyBuf, &reqdSize,
-                                NULL, NULL, NULL, NULL );
+        DWORD rc = RegEnumKeyExW(newKey, index, keyBuf, &reqdSize,
+                                 NULL, NULL, NULL, NULL );
         if (rc!=ERROR_SUCCESS) goto done;
         /* Open the sub-key */
-        rc=RegOpenKeyEx(newKey,
+        rc=RegOpenKeyExW(newKey,
                                         keyBuf, /* subkey */
                         0, /* reserved */
                         KEY_READ,
                         &subKey);
         if (rc!=ERROR_SUCCESS) goto done;
         /* Find the value of the buffer size, malloc, then read it */
-        RegQueryValueEx(subKey, NULL, 0, NULL, NULL, &reqdSize);
+        RegQueryValueExW(subKey, NULL, 0, NULL, NULL, &reqdSize);
         if (reqdSize) {
-            ppPaths[index] = malloc(reqdSize);
+            ppPaths[index] = PyMem_RawMalloc(reqdSize);
             if (ppPaths[index]) {
-                RegQueryValueEx(subKey, NULL, 0, NULL,
+                RegQueryValueExW(subKey, NULL, 0, NULL,
                                 (LPBYTE)ppPaths[index],
                                 &reqdSize);
                 dataSize += reqdSize + 1; /* 1 for the ";" */
@@ -303,19 +371,18 @@ getpythonregpath(HKEY keyBase, int skipcore)
     if (dataSize == 0) goto done;
 
     /* original datasize from RegQueryInfo doesn't include the \0 */
-    dataBuf = malloc((dataSize+1) * sizeof(TCHAR));
+    dataBuf = PyMem_RawMalloc((dataSize+1) * sizeof(WCHAR));
     if (dataBuf) {
-        TCHAR *szCur = dataBuf;
-        DWORD reqdSize = dataSize;
+        WCHAR *szCur = dataBuf;
         /* Copy our collected strings */
         for (index=0;index<numKeys;index++) {
             if (index > 0) {
-                *(szCur++) = _T(';');
+                *(szCur++) = L';';
                 dataSize--;
             }
             if (ppPaths[index]) {
-                Py_ssize_t len = _tcslen(ppPaths[index]);
-                _tcsncpy(szCur, ppPaths[index], len);
+                Py_ssize_t len = wcslen(ppPaths[index]);
+                wcsncpy(szCur, ppPaths[index], len);
                 szCur += len;
                 assert(dataSize > (DWORD)len);
                 dataSize -= (DWORD)len;
@@ -326,42 +393,32 @@ getpythonregpath(HKEY keyBase, int skipcore)
         else {
             /* If we have no values, we dont need a ';' */
             if (numKeys) {
-                *(szCur++) = _T(';');
+                *(szCur++) = L';';
                 dataSize--;
             }
             /* Now append the core path entries -
                this will include the NULL
             */
-            rc = RegQueryValueEx(newKey, NULL, 0, NULL,
-                                 (LPBYTE)szCur, &dataSize);
+            rc = RegQueryValueExW(newKey, NULL, 0, NULL,
+                                  (LPBYTE)szCur, &dataSize);
+            if (rc != ERROR_SUCCESS) {
+                PyMem_RawFree(dataBuf);
+                goto done;
+            }
         }
-        /* And set the result - caller must free
-           If MBCS, it is fine as is.  If Unicode, allocate new
-           buffer and convert.
-        */
-#ifdef UNICODE
-        retval = (char *)malloc(reqdSize+1);
-        if (retval)
-            WideCharToMultiByte(CP_ACP, 0,
-                            dataBuf, -1, /* source */
-                    retval, reqdSize+1, /* dest */
-                    NULL, NULL);
-        free(dataBuf);
-#else
+        /* And set the result - caller must free */
         retval = dataBuf;
-#endif
     }
 done:
     /* Loop freeing my temp buffers */
     if (ppPaths) {
-        for(index=0;index<numKeys;index++)
-            if (ppPaths[index]) free(ppPaths[index]);
-        free(ppPaths);
+        for(index=0; index<numKeys; index++)
+            PyMem_RawFree(ppPaths[index]);
+        PyMem_RawFree(ppPaths);
     }
     if (newKey)
         RegCloseKey(newKey);
-    if (keyBuf)
-        free(keyBuf);
+    PyMem_RawFree(keyBuf);
     return retval;
 }
 #endif /* Py_ENABLE_SHARED */
@@ -370,53 +427,14 @@ done:
 static void
 get_progpath(void)
 {
-    extern char *Py_GetProgramName(void);
-    char *path = getenv("PATH");
-    char *prog = Py_GetProgramName();
+    extern wchar_t *Py_GetProgramName(void);
+    wchar_t *path = _wgetenv(L"PATH");
+    wchar_t *prog = Py_GetProgramName();
 
-#ifdef MS_WINDOWS
-    extern HANDLE PyWin_DLLhModule;
-#ifdef UNICODE
-    WCHAR wprogpath[MAXPATHLEN+1];
-    /* Windows documents that GetModuleFileName() will "truncate",
-       but makes no mention of the null terminator.  Play it safe.
-       PLUS Windows itself defines MAX_PATH as the same, but anyway...
-    */
-#ifdef Py_ENABLE_SHARED
-    wprogpath[MAXPATHLEN]=_T('\0');
-    if (PyWin_DLLhModule &&
-        GetModuleFileName(PyWin_DLLhModule, wprogpath, MAXPATHLEN)) {
-        WideCharToMultiByte(CP_ACP, 0,
-                            wprogpath, -1,
-                            dllpath, MAXPATHLEN+1,
-                            NULL, NULL);
-    }
-#else
-    dllpath[0] = 0;
-#endif
-    wprogpath[MAXPATHLEN]=_T('\0');
-    if (GetModuleFileName(NULL, wprogpath, MAXPATHLEN)) {
-        WideCharToMultiByte(CP_ACP, 0,
-                            wprogpath, -1,
-                            progpath, MAXPATHLEN+1,
-                            NULL, NULL);
+    if (GetModuleFileNameW(NULL, progpath, MAXPATHLEN))
         return;
-    }
-#else
-    /* static init of progpath ensures final char remains \0 */
-#ifdef Py_ENABLE_SHARED
-    if (PyWin_DLLhModule)
-        if (!GetModuleFileName(PyWin_DLLhModule, dllpath, MAXPATHLEN))
-            dllpath[0] = 0;
-#else
-    dllpath[0] = 0;
-#endif
-    if (GetModuleFileName(NULL, progpath, MAXPATHLEN))
-        return;
-#endif
-#endif
     if (prog == NULL || *prog == '\0')
-        prog = "python";
+        prog = L"python";
 
     /* If there is no slash in the argv0 path, then we have to
      * assume python is on the user's $PATH, since there's no
@@ -424,24 +442,24 @@ get_progpath(void)
      * $PATH isn't exported, you lose.
      */
 #ifdef ALTSEP
-    if (strchr(prog, SEP) || strchr(prog, ALTSEP))
+    if (wcschr(prog, SEP) || wcschr(prog, ALTSEP))
 #else
-    if (strchr(prog, SEP))
+    if (wcschr(prog, SEP))
 #endif
-        strncpy(progpath, prog, MAXPATHLEN);
+        wcsncpy(progpath, prog, MAXPATHLEN);
     else if (path) {
         while (1) {
-            char *delim = strchr(path, DELIM);
+            wchar_t *delim = wcschr(path, DELIM);
 
             if (delim) {
                 size_t len = delim - path;
                 /* ensure we can't overwrite buffer */
                 len = min(MAXPATHLEN,len);
-                strncpy(progpath, path, len);
+                wcsncpy(progpath, path, len);
                 *(progpath + len) = '\0';
             }
             else
-                strncpy(progpath, path, MAXPATHLEN);
+                wcsncpy(progpath, path, MAXPATHLEN);
 
             /* join() is safe for MAXPATHLEN+1 size buffer */
             join(progpath, prog);
@@ -459,27 +477,134 @@ get_progpath(void)
         progpath[0] = '\0';
 }
 
+static int
+find_env_config_value(FILE * env_file, const wchar_t * key, wchar_t * value)
+{
+    int result = 0; /* meaning not found */
+    char buffer[MAXPATHLEN*2+1];  /* allow extra for key, '=', etc. */
+
+    fseek(env_file, 0, SEEK_SET);
+    while (!feof(env_file)) {
+        char * p = fgets(buffer, MAXPATHLEN*2, env_file);
+        wchar_t tmpbuffer[MAXPATHLEN*2+1];
+        PyObject * decoded;
+        size_t n;
+
+        if (p == NULL)
+            break;
+        n = strlen(p);
+        if (p[n - 1] != '\n') {
+            /* line has overflowed - bail */
+            break;
+        }
+        if (p[0] == '#')    /* Comment - skip */
+            continue;
+        decoded = PyUnicode_DecodeUTF8(buffer, n, "surrogateescape");
+        if (decoded != NULL) {
+            Py_ssize_t k;
+            k = PyUnicode_AsWideChar(decoded,
+                                     tmpbuffer, MAXPATHLEN * 2);
+            Py_DECREF(decoded);
+            if (k >= 0) {
+                wchar_t * context = NULL;
+                wchar_t * tok = wcstok_s(tmpbuffer, L" \t\r\n", &context);
+                if ((tok != NULL) && !wcscmp(tok, key)) {
+                    tok = wcstok_s(NULL, L" \t", &context);
+                    if ((tok != NULL) && !wcscmp(tok, L"=")) {
+                        tok = wcstok_s(NULL, L"\r\n", &context);
+                        if (tok != NULL) {
+                            wcsncpy(value, tok, MAXPATHLEN);
+                            result = 1;
+                            break;
+                        }
+                    }
+                }
+            }
+        }
+    }
+    return result;
+}
+
 static void
 calculate_path(void)
 {
-    char argv0_path[MAXPATHLEN+1];
-    char *buf;
+    wchar_t argv0_path[MAXPATHLEN+1];
+    wchar_t *buf;
     size_t bufsz;
-    char *pythonhome = Py_GetPythonHome();
-    char *envpath = Py_GETENV("PYTHONPATH");
+    wchar_t *pythonhome = Py_GetPythonHome();
+    wchar_t *envpath = NULL;
 
 #ifdef MS_WINDOWS
     int skiphome, skipdefault;
-    char *machinepath = NULL;
-    char *userpath = NULL;
-    char zip_path[MAXPATHLEN+1];
-    size_t len;
+    wchar_t *machinepath = NULL;
+    wchar_t *userpath = NULL;
+    wchar_t zip_path[MAXPATHLEN+1];
+    int applocal = 0;
+
+    if (!Py_IgnoreEnvironmentFlag) {
+        envpath = _wgetenv(L"PYTHONPATH");
+    }
+#else
+    char *_envpath = Py_GETENV("PYTHONPATH");
+    wchar_t wenvpath[MAXPATHLEN+1];
+    if (_envpath) {
+        size_t r = mbstowcs(wenvpath, _envpath, MAXPATHLEN+1);
+        envpath = wenvpath;
+        if (r == (size_t)-1 || r >= MAXPATHLEN)
+            envpath = NULL;
+    }
 #endif
 
     get_progpath();
     /* progpath guaranteed \0 terminated in MAXPATH+1 bytes. */
-    strcpy(argv0_path, progpath);
+    wcscpy_s(argv0_path, MAXPATHLEN+1, progpath);
     reduce(argv0_path);
+
+    /* Search for an environment configuration file, first in the
+       executable's directory and then in the parent directory.
+       If found, open it for use when searching for prefixes.
+    */
+
+    {
+        wchar_t envbuffer[MAXPATHLEN+1];
+        wchar_t tmpbuffer[MAXPATHLEN+1];
+        const wchar_t *env_cfg = L"pyvenv.cfg";
+        FILE * env_file = NULL;
+
+        wcscpy_s(envbuffer, MAXPATHLEN+1, argv0_path);
+        join(envbuffer, env_cfg);
+        env_file = _Py_wfopen(envbuffer, L"r");
+        if (env_file == NULL) {
+            errno = 0;
+            reduce(envbuffer);
+            reduce(envbuffer);
+            join(envbuffer, env_cfg);
+            env_file = _Py_wfopen(envbuffer, L"r");
+            if (env_file == NULL) {
+                errno = 0;
+            }
+        }
+        if (env_file != NULL) {
+            /* Look for an 'applocal' variable and, if true, ignore all registry
+             * keys and environment variables, but retain the default paths
+             * (DLLs, Lib) and the zip file. Setting pythonhome here suppresses
+             * the search for LANDMARK below and overrides %PYTHONHOME%.
+             */
+            if (find_env_config_value(env_file, L"applocal", tmpbuffer) &&
+                (applocal = (wcsicmp(tmpbuffer, L"true") == 0))) {
+                envpath = NULL;
+                pythonhome = argv0_path;
+            }
+            
+            /* Look for a 'home' variable and set argv0_path to it, if found */
+            if (find_env_config_value(env_file, L"home", tmpbuffer)) {
+                wcscpy_s(argv0_path, MAXPATHLEN+1, tmpbuffer);
+            }
+            fclose(env_file);
+            env_file = NULL;
+        }
+    }
+
     if (pythonhome == NULL || *pythonhome == '\0') {
         if (search_for_prefix(argv0_path, LANDMARK))
             pythonhome = prefix;
@@ -487,33 +612,26 @@ calculate_path(void)
             pythonhome = NULL;
     }
     else
-        strncpy(prefix, pythonhome, MAXPATHLEN);
+        wcscpy_s(prefix, MAXPATHLEN+1, pythonhome);
 
     if (envpath && *envpath == '\0')
         envpath = NULL;
 
 
 #ifdef MS_WINDOWS
-    /* Calculate zip archive path */
-    if (dllpath[0])             /* use name of python DLL */
-        strncpy(zip_path, dllpath, MAXPATHLEN);
-    else                        /* use name of executable program */
-        strncpy(zip_path, progpath, MAXPATHLEN);
-    zip_path[MAXPATHLEN] = '\0';
-    len = strlen(zip_path);
-    if (len > 4) {
-        zip_path[len-3] = 'z';          /* change ending to "zip" */
-        zip_path[len-2] = 'i';
-        zip_path[len-1] = 'p';
-    }
-    else {
-        zip_path[0] = 0;
+    /* Calculate zip archive path from DLL or exe path */
+    if (!get_dllpath(zip_path)) {
+        change_ext(zip_path, zip_path, L".zip");
+    } else {
+        change_ext(zip_path, progpath, L".zip");
     }
 
     skiphome = pythonhome==NULL ? 0 : 1;
 #ifdef Py_ENABLE_SHARED
-    machinepath = getpythonregpath(HKEY_LOCAL_MACHINE, skiphome);
-    userpath = getpythonregpath(HKEY_CURRENT_USER, skiphome);
+    if (!applocal) {
+        machinepath = getpythonregpath(HKEY_LOCAL_MACHINE, skiphome);
+        userpath = getpythonregpath(HKEY_CURRENT_USER, skiphome);
+    }
 #endif
     /* We only use the default relative PYTHONPATH if we havent
        anything better to use! */
@@ -532,33 +650,34 @@ calculate_path(void)
        Extra rules:
        - If PYTHONHOME is set (in any way) item (3) is ignored.
        - If registry values are used, (4) and (5) are ignored.
+       - If applocal is set, (1), (3), and registry values are ignored
     */
 
     /* Calculate size of return buffer */
     if (pythonhome != NULL) {
-        char *p;
+        wchar_t *p;
         bufsz = 1;
         for (p = PYTHONPATH; *p; p++) {
             if (*p == DELIM)
                 bufsz++; /* number of DELIM plus one */
         }
-        bufsz *= strlen(pythonhome);
+        bufsz *= wcslen(pythonhome);
     }
     else
         bufsz = 0;
-    bufsz += strlen(PYTHONPATH) + 1;
-    bufsz += strlen(argv0_path) + 1;
+    bufsz += wcslen(PYTHONPATH) + 1;
+    bufsz += wcslen(argv0_path) + 1;
 #ifdef MS_WINDOWS
-    if (userpath)
-        bufsz += strlen(userpath) + 1;
-    if (machinepath)
-        bufsz += strlen(machinepath) + 1;
-    bufsz += strlen(zip_path) + 1;
+    if (!applocal && userpath)
+        bufsz += wcslen(userpath) + 1;
+    if (!applocal && machinepath)
+        bufsz += wcslen(machinepath) + 1;
+    bufsz += wcslen(zip_path) + 1;
 #endif
     if (envpath != NULL)
-        bufsz += strlen(envpath) + 1;
+        bufsz += wcslen(envpath) + 1;
 
-    module_search_path = buf = malloc(bufsz);
+    module_search_path = buf = PyMem_RawMalloc(bufsz*sizeof(wchar_t));
     if (buf == NULL) {
         /* We can't exit, so print a warning and limp along */
         fprintf(stderr, "Can't malloc dynamic PYTHONPATH.\n");
@@ -571,79 +690,85 @@ calculate_path(void)
             module_search_path = PYTHONPATH;
         }
 #ifdef MS_WINDOWS
-        if (machinepath)
-            free(machinepath);
-        if (userpath)
-            free(userpath);
+        PyMem_RawFree(machinepath);
+        PyMem_RawFree(userpath);
 #endif /* MS_WINDOWS */
         return;
     }
 
     if (envpath) {
-        strcpy(buf, envpath);
-        buf = strchr(buf, '\0');
+        if (wcscpy_s(buf, bufsz - (buf - module_search_path), envpath))
+            Py_FatalError("buffer overflow in getpathp.c's calculate_path()");
+        buf = wcschr(buf, L'\0');
         *buf++ = DELIM;
     }
 #ifdef MS_WINDOWS
     if (zip_path[0]) {
-        strcpy(buf, zip_path);
-        buf = strchr(buf, '\0');
+        if (wcscpy_s(buf, bufsz - (buf - module_search_path), zip_path))
+            Py_FatalError("buffer overflow in getpathp.c's calculate_path()");
+        buf = wcschr(buf, L'\0');
         *buf++ = DELIM;
     }
     if (userpath) {
-        strcpy(buf, userpath);
-        buf = strchr(buf, '\0');
+        if (wcscpy_s(buf, bufsz - (buf - module_search_path), userpath))
+            Py_FatalError("buffer overflow in getpathp.c's calculate_path()");
+        buf = wcschr(buf, L'\0');
         *buf++ = DELIM;
-        free(userpath);
+        PyMem_RawFree(userpath);
     }
     if (machinepath) {
-        strcpy(buf, machinepath);
-        buf = strchr(buf, '\0');
+        if (wcscpy_s(buf, bufsz - (buf - module_search_path), machinepath))
+            Py_FatalError("buffer overflow in getpathp.c's calculate_path()");
+        buf = wcschr(buf, L'\0');
         *buf++ = DELIM;
-        free(machinepath);
+        PyMem_RawFree(machinepath);
     }
     if (pythonhome == NULL) {
         if (!skipdefault) {
-            strcpy(buf, PYTHONPATH);
-            buf = strchr(buf, '\0');
+            if (wcscpy_s(buf, bufsz - (buf - module_search_path), PYTHONPATH))
+                Py_FatalError("buffer overflow in getpathp.c's calculate_path()");
+            buf = wcschr(buf, L'\0');
+            *buf++ = DELIM;
         }
     }
 #else
     if (pythonhome == NULL) {
-        strcpy(buf, PYTHONPATH);
-        buf = strchr(buf, '\0');
+        wcscpy(buf, PYTHONPATH);
+        buf = wcschr(buf, L'\0');
+        *buf++ = DELIM;
     }
 #endif /* MS_WINDOWS */
     else {
-        char *p = PYTHONPATH;
-        char *q;
+        wchar_t *p = PYTHONPATH;
+        wchar_t *q;
         size_t n;
         for (;;) {
-            q = strchr(p, DELIM);
+            q = wcschr(p, DELIM);
             if (q == NULL)
-                n = strlen(p);
+                n = wcslen(p);
             else
                 n = q-p;
             if (p[0] == '.' && is_sep(p[1])) {
-                strcpy(buf, pythonhome);
-                buf = strchr(buf, '\0');
+                if (wcscpy_s(buf, bufsz - (buf - module_search_path), pythonhome))
+                    Py_FatalError("buffer overflow in getpathp.c's calculate_path()");
+                buf = wcschr(buf, L'\0');
                 p++;
                 n--;
             }
-            strncpy(buf, p, n);
+            wcsncpy(buf, p, n);
             buf += n;
+            *buf++ = DELIM;
             if (q == NULL)
                 break;
-            *buf++ = DELIM;
             p = q+1;
         }
     }
     if (argv0_path) {
+        wcscpy(buf, argv0_path);
+        buf = wcschr(buf, L'\0');
         *buf++ = DELIM;
-        strcpy(buf, argv0_path);
-        buf = strchr(buf, '\0');
     }
-    *buf = '\0';
+    *(buf - 1) = L'\0';
     /* Now to pull one last hack/trick.  If sys.prefix is
        empty, then try and find it somewhere on the paths
        we calculated.  We scan backwards, as our general policy
@@ -652,12 +777,12 @@ calculate_path(void)
        on the path, and that our 'prefix' directory is
        the parent of that.
     */
-    if (*prefix=='\0') {
-        char lookBuf[MAXPATHLEN+1];
-        char *look = buf - 1; /* 'buf' is at the end of the buffer */
+    if (*prefix==L'\0') {
+        wchar_t lookBuf[MAXPATHLEN+1];
+        wchar_t *look = buf - 1; /* 'buf' is at the end of the buffer */
         while (1) {
             Py_ssize_t nchars;
-            char *lookEnd = look;
+            wchar_t *lookEnd = look;
             /* 'look' will end up one character before the
                start of the path in question - even if this
                is one character before the start of the buffer
@@ -665,8 +790,8 @@ calculate_path(void)
             while (look >= module_search_path && *look != DELIM)
                 look--;
             nchars = lookEnd-look;
-            strncpy(lookBuf, look+1, nchars);
-            lookBuf[nchars] = '\0';
+            wcsncpy(lookBuf, look+1, nchars);
+            lookBuf[nchars] = L'\0';
             /* Up one level to the parent */
             reduce(lookBuf);
             if (search_for_prefix(lookBuf, LANDMARK)) {
@@ -681,9 +806,25 @@ calculate_path(void)
 }
 
 
-/* External interface */
+void
+Py_SetPath(const wchar_t *path)
+{
+    if (module_search_path != NULL) {
+        PyMem_RawFree(module_search_path);
+        module_search_path = NULL;
+    }
+    if (path != NULL) {
+        extern wchar_t *Py_GetProgramName(void);
+        wchar_t *prog = Py_GetProgramName();
+        wcsncpy(progpath, prog, MAXPATHLEN);
+        prefix[0] = L'\0';
+        module_search_path = PyMem_RawMalloc((wcslen(path) + 1) * sizeof(wchar_t));
+        if (module_search_path != NULL)
+            wcscpy(module_search_path, path);
+    }
+}
 
-char *
+wchar_t *
 Py_GetPath(void)
 {
     if (!module_search_path)
@@ -691,7 +832,7 @@ Py_GetPath(void)
     return module_search_path;
 }
 
-char *
+wchar_t *
 Py_GetPrefix(void)
 {
     if (!module_search_path)
@@ -699,16 +840,65 @@ Py_GetPrefix(void)
     return prefix;
 }
 
-char *
+wchar_t *
 Py_GetExecPrefix(void)
 {
     return Py_GetPrefix();
 }
 
-char *
+wchar_t *
 Py_GetProgramFullPath(void)
 {
     if (!module_search_path)
         calculate_path();
     return progpath;
 }
+
+/* Load python3.dll before loading any extension module that might refer
+   to it. That way, we can be sure that always the python3.dll corresponding
+   to this python DLL is loaded, not a python3.dll that might be on the path
+   by chance.
+   Return whether the DLL was found.
+*/
+static int python3_checked = 0;
+static HANDLE hPython3;
+int
+_Py_CheckPython3()
+{
+    wchar_t py3path[MAXPATHLEN+1];
+    if (python3_checked) {
+        return hPython3 != NULL;
+    }
+    python3_checked = 1;
+
+    /* If there is a python3.dll next to the python3y.dll,
+       use that DLL */
+    if (!get_dllpath(py3path)) {
+        reduce(py3path);
+        join(py3path, PY3_DLLNAME);
+        hPython3 = LoadLibraryExW(py3path, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
+        if (hPython3 != NULL) {
+            return 1;
+        }
+    }
+
+    /* If we can locate python3.dll in our application dir,
+       use that DLL */
+    wcscpy(py3path, Py_GetPrefix());
+    if (py3path[0]) {
+        join(py3path, PY3_DLLNAME);
+        hPython3 = LoadLibraryExW(py3path, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
+        if (hPython3 != NULL) {
+            return 1;
+        }
+    }
+
+    /* For back-compat, also search {sys.prefix}\DLLs, though
+       that has not been a normal install layout for a while */
+    wcscpy(py3path, Py_GetPrefix());
+    if (py3path[0]) {
+        join(py3path, L"DLLs\\" PY3_DLLNAME);
+        hPython3 = LoadLibraryExW(py3path, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);
+    }
+    return hPython3 != NULL;
+}
